class Packet:
    ...


class Request(Packet):
    ...


class FrameRequest(Request):
    def __init__(self, framesize):
        self.framesize = framesize


class Frame(Packet):
    def __init__(self, frame):
        self.frame = frame


class LastFrame(Frame):
    def __init__(self, frame, valid_samples):
        self.frame = frame
        self.valid_samples = valid_samples


class TriggerFrame(Frame):
    def __init__(self, frame, indices):
        self.frame = frame
        self.indices = indices


class GateOpenFrame(Frame):
    ...


class GateCloseFrame(Frame):
    ...


class GateOpenCloseFrame(GateOpenFrame, GateCloseFrame):
    ...


class Pipeline:
    def __init__(self, *nodes):
        self.nodes = nodes
        for node in self:
            node._pipeline = self

    def __str__(self):
        return str([node.__class__.__name__ for node in self])

    def __iter__(self):
        return iter(self.nodes)

    def to_dict(self):
        from . import __version_info__
        return dict(
            package_version=__version_info__,
            nodes=[node.to_dict() for node in self]
        )

    @property
    def samplerate_decider(self):
        for node in self:
            if isinstance(node, SamplerateDecider):
                return node

    def run(self):
        self.samplerate_decider.run()

    def setup(self):
        for node in self:
            node.setup()

    def reset(self):
        for node in self:
            node.reset()

    def __or__(self, other):
        if isinstance(other, Pipeline):
            self.nodes[-1]._downstream = other.nodes[0]
            other.nodes[0]._upstream = self.nodes[-1]
            return Pipeline(*self.nodes, *other.nodes)
        if isinstance(other, Node):
            self.nodes[-1]._downstream = other
            other._upstream = self.nodes[-1]
            return Pipeline(*self.nodes, other)
        return NotImplemented

    def __ror__(self, other):
        if isinstance(other, Pipeline):
            self.nodes[0]._upstream = other.nodes[-1]
            other.nodes[-1]._downstream = self.nodes[0]
            return Pipeline(*other.nodes, *self.nodes)
        if isinstance(other, Node):
            self.nodes[0]._upstream = other
            other._downstream = self.nodes[0]
            return Pipeline(other, *self.nodes)
        return NotImplemented


class Node:
    """Generic pipeline Node."""

    def __init__(self):
        self._pipeline = None
        self._is_ready = False
        self.__upstream = None
        self.__downstream = None

    def __or__(self, other):
        if isinstance(other, Node):
            self._downstream = other
            other._upstream = self
            return Pipeline(self, other)
        return NotImplemented

    def to_dict(self):
        return dict(type=self.__class__.__name__)

    def setup(self):
        """Run setup for this Node."""
        self._is_ready = True

    def reset(self):
        """Reset this Node."""
        self._is_ready = False

    def process(self, frame):
        """Process one frame of data."""
        return frame

    @property
    def _upstream(self):
        """Wrap for the upstream node object."""
        return self.__upstream

    @_upstream.setter
    def _upstream(self, node):
        """Set the upstream node, running update code."""
        self.__upstream = node
        self._upstream_changed()

    def _upstream_changed(self):
        """Code to run when the upstream is changed."""
        pass

    @property
    def _downstream(self):
        """Wrap for the downstream node object."""
        return self.__downstream

    @_downstream.setter
    def _downstream(self, node):
        """Set the downstream node, running update code."""
        self.__downstream = node
        self._downstream_changed()

    def _downstream_changed(self):
        """Code to run when the downstream is changed."""
        pass

    def push(self, packet):
        """Give a packet of data as the input to this node.

        Processes a frame and passes it along down the pipeline.
        """
        if isinstance(packet, Frame):
            packet = self.process(packet)

        if packet is None:
            return

        if self._downstream is not None:
            # We should push the frame down the pipeline, and return the final result.
            return self._downstream.push(packet)
        else:
            # This is the end of the pipeline, so return the frame.
            return packet

    def request(self, packet):
        """Request one frame of output from this object.

        Requests a frame from the upstream or this node, process it,
        then return the frame.
        """
        if self._upstream is not None:
            # There's an upstream node, ask it to handle the packet first.
            packet = self._upstream.request(packet)

        if isinstance(packet, Request):
            if isinstance(packet, FrameRequest):
                # Either there's no upstream node, or it could not handle the request.
                # Process this request here, then return the generated frame.
                return self.process(packet)


class SamplerateDecider(Node):
    """Base class for nodes that have an externally defined samplerate.

    Instances of this type will decide the samplerate for an entire pipeline,
    by telling the SamplerateFollowers what samplerate they are operating at.
    """
    def __init__(self, samplerate, **kwargs):
        super().__init__(**kwargs)
        self.samplerate = samplerate

    def to_dict(self):
        return super().to_dict() | dict(
            samplerate=self.samplerate,
        )

    def _upstream_changed(self):
        super()._upstream_changed()
        if isinstance(self._upstream, SamplerateFollower):
            self._upstream._set_samplerate_direction('downstream')

    def _downstream_changed(self):
        super()._downstream_changed()
        if isinstance(self._downstream, SamplerateFollower):
            self._downstream._set_samplerate_direction('upstream')


class SamplerateFollower(Node):
    """Base class for nodes which have no external samplerate.

    Instances of this class will look for a samplerate elsewhere in the pipeline.
    This can be found either upstream when the instance is receiving data from
    a SamperateDecider, or downstream when the instance will send data to a SamplerateDecider.
    This is managed automatically when assembling the pipeline, no user interaction required.
    """
    def __init__(self, **kwargs):
        self._samplerate_direction = None
        super().__init__(**kwargs)

    def _upstream_changed(self):
        """Clear any stored samplerate direction since it might no longer be valid."""
        super()._upstream_changed()
        self._samplerate_direction = None

    def _downstream_changed(self):
        """Clear any stored samplerate direction since it might no longer be valid."""
        super()._downstream_changed()
        self._samplerate_direction = None

    @property
    def samplerate(self):
        direction = self._get_samplerate_direction()
        if direction == 'upstream':
            samplerate = self._upstream.samplerate
            try:
                _, samplerate = samplerate
            except TypeError:
                pass
            return samplerate
        if direction == 'downstream':
            samplerate = self._downstream.samplerate
            try:
                samplerate, _ = samplerate
            except TypeError:
                pass
            return samplerate

    def _get_samplerate_direction(self, direction='both'):
        """Check if the samplerate can be found along the pipeline.

        This will run along the pipeline and see if there's a SamplerateDecider
        somewhere. If a direction is specified, it will only look in the given direction.
        If found, all other SamplerateFollowers will be updated.

        This is a complicated getter for the samplerate direction, parameterized with a
        search direction to allow recursive searches. Additionally, when the value is found
        it will be updated for all the relevant SamplerateFollowers.
        """
        if self._samplerate_direction is not None:
            return self._samplerate_direction

        if direction == 'both':
            self._samplerate_direction = self._get_samplerate_direction('upstream') or self._get_samplerate_direction('downstream')
            return self._samplerate_direction

        if direction == 'upstream':
            if isinstance(self._upstream, SamplerateDecider):
                self._set_samplerate_direction('upstream')
                return 'upstream'
            elif isinstance(self._upstream, SamplerateFollower):
                return self._upstream._get_samplerate_direction('upstream')
        elif direction == 'downstream':
            if isinstance(self._downstream, SamplerateDecider):
                self._set_samplerate_direction('downstream')
                return 'downstream'
            elif isinstance(self._downstream, SamplerateFollower):
                return self._downstream._get_samplerate_direction('downstream')
        else:
            raise ValueError(f"Search direction should be one of 'upstream' or 'downstream', got {direction}")
        return False

    def _set_samplerate_direction(self, direction):
        """Update the pipeline samplerate direction.

        This will go through the pipeline in one direction and update
        the samplerate search direction.
        It will always go through the pipeline in the opposite direction to the
        search direction. I.e., if diretcion='upstream', this will go downstream
        set all the samplerate directions to 'upstream', and vice versa.
        """
        if direction == 'upstream':
            self._samplerate_direction = 'upstream'
            if isinstance(self._downstream, SamplerateFollower):
                if self._downstream._samplerate_direction is None:
                    self._downstream._set_samplerate_direction('upstream')
                elif self._downstream._samplerate_direction == 'downstream':
                    raise ValueError('Conflicting pipeline samplerates!')
        elif direction == 'downstream':
            self._samplerate_direction = 'downstream'
            if isinstance(self._upstream, SamplerateFollower):
                if self._upstream._samplerate_direction is None:
                    self._upstream._set_samplerate_direction('downstream')
                elif self._upstream._samplerate_direction == 'upstream':
                    raise ValueError('Conflicting pipeline samplerates!')



from . import _core
from . import timestamps
import numpy as np
import zarr
import threading
import queue
import os


class _Recorder(_core.SamplerateFollower):
    ...


class InternalRecorder(_Recorder):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.reset()

    def reset(self, **kwargs):
        super().reset(**kwargs)
        self._storage = [[]]

    def process(self, frame):
        if isinstance(frame, _core.Frame):
            self._storage[-1].append(frame)
            if isinstance(frame, _core.GateCloseFrame):
                self._storage.append([])
        return frame

    @property
    def recorded_data(self):
        series = [np.concatenate([frame.frame for frame in series], axis=-1) for series in self._storage if len(series) > 0]
        if len(series) == 1:
            return series[0]
        try:
            return np.stack(series, axis=0)
        except ValueError:
            return series


class ZArrRecorder(_Recorder):
    timeout = 1
    def __init__(
        self,
        filename=None,
        mode='w-',
        chunksize=50000,
        threaded=False,
        **kwargs
    ):
        super().__init__(**kwargs)
        if filename is None:
            filename = timestamps.timestamp('filename')
        base, ext = os.path.splitext(filename)
        if ext == '':
            ext = '.zarr'
        self.filename = base + ext
        self.mode = mode
        if self.mode == 'w-' and os.path.exists(self.filename):
            raise FileExistsError(f"File {self.filename} already exists on disk, use overwriting ('w') or appending ('a') mode if desired.")
        self.chunksize = chunksize
        self.threaded = threaded

    def to_dict(self):
        return super().to_dict() | dict(
            filename=self.filename,
            chunksize=self.chunksize,
            thereaded=self.threaded,
            mode=self.mode,
        )

    def setup(self, **kwargs):
        super().setup(**kwargs)
        if self.threaded:
            self._q = queue.Queue()
            self._thread = threading.Thread(target=self.threaded_writer)
            self._stop_event = threading.Event()
            self._thread.start()
        else:
            self.writer = self.initialize_writer()
            self.writer.send(None)

    def reset(self, **kwargs):
        super().reset(**kwargs)
        self._is_ready = False

    def process(self, frame):
        if not self._is_ready:
            self.setup()
        if isinstance(frame, _core.Frame):
            if self.threaded:
                self._q.put(frame)
            else:
                try:
                    self.writer.send(frame)
                except StopIteration:
                    pass
        return frame

    def initialize_writer(self):
        self.store = zarr.storage.DirectoryStore(self.filename)

        frame = (yield)
        channels, framesize = frame.frame.shape
        chunksize = self.chunksize

        buffer = np.zeros((channels, chunksize))
        self.data = zarr.open_array(
            store=self.store,
            mode=self.mode,
            shape=(0, channels, 0),
            chunks=(1, 1, chunksize),
        )
        self.data.attrs['samplerate'] = self.samplerate
        self.data.attrs['pipeline'] = self._pipeline.to_dict()
        buffer_idx = 0
        shot_idx = -1
        while not isinstance(frame, _core.LastFrame):
            if frame is None:
                frame = (yield)
            channels, framesize = frame.frame.shape

            if isinstance(frame, _core.GateOpenFrame) or shot_idx == -1:
                assert buffer_idx == 0, "There seems to be data in the buffer at the start of a shot recording!"
                shot_idx += 1
                _shots, _channels, _samples = self.data.shape
                self.data.resize((_shots + 1, _channels, _samples))
                write_idx = 0

            if framesize < chunksize - buffer_idx:
                # This frame can fit in the buffer
                buffer[:, buffer_idx:buffer_idx + framesize] = frame.frame
                buffer_idx += framesize
                if isinstance(frame, _core.GateCloseFrame):
                    # Gate just closed, we should write the frame and then get a new one
                    frame = None
                else:
                    # In the middle of the open segment, just get the next frame without writing to file
                    frame = None
                    continue
            else:
                # The frame will fill the buffer
                buffer[:, buffer_idx:] = frame.frame[:, :chunksize - buffer_idx]
                if chunksize - buffer_idx < framesize:
                    # The remainder is kept to next loop iteration
                    frame.frame = frame.frame[:, chunksize - buffer_idx:]
                else:
                    # The frame perfectly fit in the buffer, so we need a new one the next iteration
                    frame = None
                buffer_idx = chunksize

            # The buffer is full, time to write to file
            if write_idx + buffer_idx >= self.data.shape[-1]:
                # Array is too small to hold the entire shot
                _shots, _channels, _ = self.data.shape
                _samples = write_idx + buffer_idx
                _samples = _samples + (-_samples % chunksize)
                self.data.resize((_shots, _channels, _samples))
            self.data[shot_idx, :, write_idx:write_idx + buffer_idx] = buffer[:, :buffer_idx]
            write_idx += buffer_idx
            buffer_idx = 0

    def zip_store(self):
        zipname = self.filename + '.zip'
        if os.path.exists(zipname):
            if self.mode == 'w':
                os.remove(zipname)
            else:
                raise FileExistsError(f"Cannot zip data to existing file '{zipname}'. Merging recordings is not supported.")
        zip_store = zarr.storage.ZipStore(
            zipname,
            compression=zarr.storage.zipfile.ZIP_DEFLATED,
        )
        zarr.copy_store(self.store, zip_store)
        zarr.storage.rmdir(self.store)
        self.store = zip_store
        self.data = zarr.open_array(self.store)

    def threaded_writer(self):
        writer = self.initialize_writer()
        writer.send(None)

        while not self._stop_event.is_set():
            try:
                frame = self._q.get(timeout=self.timeout)
            except queue.Empty:
                # Timeout, proceed to check the stop event
                continue
            if frame is None:
                # This signals a stop
                self._stop_event.set()
                continue
            writer.send(frame)
        try:
            writer.send(None)
        except StopIteration:
            pass



import numpy as np
import fractions


def _apply_to_signals(func, signals):
    if isinstance(signals, np.ndarray):
        return func(signals)
    try:
        return tuple(func(signal) for signal in signals)
    except TypeError as err:
        if str(err).endswith('object is not iterable'):
            return func(signals)
        else:
            raise


def _length_parser(length, samplerate):
    if isinstance(length, str):
        if '%' in length:
            length = float(length.strip('%')) / 100
            def parser(signal):
                return round(signal.shape[-1] * length)
        else:
            length = fractions.Fraction(length)
            def parser(signal):
                return round(signal.shape[-1] * length)
    else:
        length = length or 0
        if samplerate is not None:
            length = round(samplerate * length)
        def parser(signal):
            return length
    return parser


def truncate_signals(signals, length=None, samplerate=None):
    """Cut signals down to a specified length."""
    if length is None:
        return signals
    if samplerate is not None:
        length = round(samplerate * length)

    def truncation(signal):
        return signal[..., :length]

    return _apply_to_signals(truncation, signals)


def extend_signals(signals, length=None, samplerate=None):
    """Zero-pad signals to a certain length.

    Note that this will not shorten signals, so if they are longer
    than the requested length they will be unchanged.
    """
    if length is None:
        return signals
    if samplerate is not None:
        length = round(samplerate * length)

    def extend(signal):
        padding = length - signal.shape[-1]
        if padding < 1:
            return signal.copy()
        padding = np.zeros(signal.shape[:-1] + (padding,))
        padded = np.concatenate([signal, padding], axis=-1)
        return padded

    return _apply_to_signals(extend, signals)


def pad_signals(signals, pre_pad=None, post_pad=None, samplerate=None):
    if pre_pad is None and post_pad is None:
        return signals

    pre_pad = _length_parser(pre_pad, samplerate)
    post_pad = _length_parser(post_pad, samplerate)
    def pad(signal):
        pre = np.zeros(signal.shape[:-1] + (pre_pad(signal),))
        post = np.zeros(signal.shape[:-1] + (post_pad(signal),))
        return np.concatenate([pre, signal, post], axis=-1)

    return _apply_to_signals(pad, signals)


def fade_signals(signals, fade_in=None, fade_out=None, samplerate=None, inplace=False):
    if fade_in is None and fade_out is None:
        return signals

    fade_in = _length_parser(fade_in, samplerate)
    fade_out = _length_parser(fade_out, samplerate)

    def fade(signal):
        if not inplace:
            signal = signal.copy()
        in_samples = fade_in(signal)
        out_samples = fade_out(signal)
        signal[..., :in_samples] *= np.sin(np.linspace(0, np.pi / 2, in_samples))**2
        signal[..., -out_samples:] *= np.sin(np.linspace(np.pi / 2, 0, out_samples))**2
        return signal

    return _apply_to_signals(fade, signals)


def nonzero_signals(signals, inplace=False):
    def nonzero(signal):
        if not inplace:
            signal = signal.copy()
        zero_indices = signal == 0
        signal[zero_indices] = np.min(np.abs(signal[np.invert(zero_indices)]))
        return signal
    return _apply_to_signals(nonzero, signals)



import numpy as np
import scipy.signal
from . import _core
import collections


class Gate(_core.SamplerateFollower):
    def __init__(
        self,
        open_trigger=None,
        close_trigger=None,
        pre_trigger=None,
        post_trigger=None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.open_trigger = open_trigger
        self.close_trigger = close_trigger
        self.pre_trigger = pre_trigger
        self.post_trigger = post_trigger

    def to_dict(self):
        try:
            open_trigger = self.open_trigger.to_dict()
        except AttributeError:
            open_trigger = str(self.open_trigger)
        try:
            close_trigger = self.close_trigger.to_dict()
        except AttributeError:
            close_trigger = str(self.close_trigger)
        return super().to_dict() | dict(
            open_trigger=open_trigger,
            close_trigger=close_trigger,
            pre_trigger=self.pre_trigger,
            post_trigger=self.post_trigger,
        )

    def setup(self, **kwargs):
        super().setup(**kwargs)
        if self.open_trigger is not None:
            self.open_trigger.setup()
        if self.close_trigger is not None:
            self.close_trigger.setup()
        if self.pre_trigger:
            self._pre_trigger_samples = np.math.ceil(self.pre_trigger * self.samplerate)
            self._pre_trigger_buffer = collections.deque()
        if self.post_trigger:
            self._post_trigger_samples = np.math.ceil(self.post_trigger * self.samplerate)
        self.reset()
        #     self._post_trigger_buffer = collections.deque()
        #     self._kept_samples = 0

    def reset(self, **kwargs):
        super().reset(**kwargs)
        self.open = True if self.open_trigger is None else False
        self.waiting = True
        self._post_close_buffer = None
        if self.open_trigger is not None:
            self.open_trigger.reset()
        if self.close_trigger is not None:
            self.close_trigger.reset()
        if self.pre_trigger:
            self._pre_trigger_buffer.clear()
            self._buffered_samples = 0
        if self.post_trigger:
            self._remaining_post_trigger_samples = 0

    @property
    def open_trigger(self):
        return self._open_trigger

    @open_trigger.setter
    def open_trigger(self, trigger):
        if trigger is not None:
            self._open_trigger = trigger
            trigger._upstream = self
        else:
            self._open_trigger = None

    @property
    def close_trigger(self):
        return self._close_trigger

    @close_trigger.setter
    def close_trigger(self, trigger):
        if trigger is not None:
            self._close_trigger = trigger
            trigger._upstream = self
        else:
            self._close_trigger = None

    def process(self, frame):
        # The gate can be in the following states, which happen in this order
        # 1) Closed and waiting for the opening trigger
        # 2) Open and waiting for the closing trigger
        # 3) Open and collecting post-trigger data
        # 4) Closed with no intent of opening?
        if self._post_close_buffer is not None:
            frame.frame = np.concatenate((self._post_close_buffer, frame.frame), axis=1)
            self._post_close_buffer = None
        if not self.open and self.waiting and self.open_trigger is not None:
            return self._wait_for_opening_trigger(frame)
        if self.open and self.waiting and self.close_trigger is not None:
            return self._wait_for_closing_trigger(frame)
        if self.open and not self.waiting:
            return self._collect_post_trigger_samples(frame)

    def _wait_for_opening_trigger(self, frame):
        frame = self.open_trigger.process(frame)
        if isinstance(frame, _core.TriggerFrame):
            trig_idx = frame.indices[0]
            frame.frame, post_frame = frame.frame[:, :trig_idx], frame.frame[:, trig_idx:]
        else:
            trig_idx = None

        if self.pre_trigger:
            self._pre_trigger_buffer.append(frame.frame)
            self._buffered_samples += frame.frame.shape[1]
            while self._buffered_samples - self._pre_trigger_buffer[0].shape[1] > self._pre_trigger_samples:
                self._buffered_samples -= self._pre_trigger_buffer.popleft().shape[1]

        if trig_idx is None:
            # Gate is stil closed!
            return

        self.open = True
        self.open_trigger.reset()  # We won't feed data into the open trigger for a while, so it needs to start anew on next call.
        if self.pre_trigger:
            pre_frame = np.concatenate(self._pre_trigger_buffer, axis=1)[:, -self._pre_trigger_samples:]
            self._pre_trigger_buffer.clear()
            self._buffered_samples = 0
        else:
            pre_frame = np.zeros((frame.frame.shape[0], 0))
        # frame = _core.GateOpenFrame(frame.frame[:, ])

        # Run the close trigger on the frame from the opening trigger point, not including the pre-trigger data.
        if self.close_trigger is not None:
            self.waiting = True
            post_frame = self._wait_for_closing_trigger(_core.Frame(post_frame))
        elif self.post_trigger:
            self.waiting = False
            self._remaining_post_trigger_samples = self._post_trigger_samples + 1
            post_frame = self._collect_post_trigger_samples(_core.Frame(post_frame))
        else:
            post_frame = _core.GateCloseFrame(post_frame[:, :1])

        frame = np.concatenate((pre_frame, post_frame.frame), axis=1)
        if isinstance(post_frame, _core.GateCloseFrame):
            return _core.GateOpenCloseFrame(frame)
        return _core.GateOpenFrame(frame)

    def _wait_for_closing_trigger(self, frame):
        frame = self.close_trigger.process(frame)
        if not isinstance(frame, _core.TriggerFrame):
            return frame

        trig_idx = frame.indices[0] + 1
        self.close_trigger.reset()  # We won't feed data into the close trigger for a while, so it needs to start anew on next call.
        pre_frame, post_frame = frame.frame[:, :trig_idx], frame.frame[:, trig_idx:]
        if self.post_trigger:
            self.waiting = False
            self._remaining_post_trigger_samples = self._post_trigger_samples
            post_frame = self._collect_post_trigger_samples(_core.Frame(post_frame))
            post_frame.frame = np.concatenate([pre_frame, post_frame.frame], axis=1)
            return post_frame
        else:
            self.open = False
            self._post_close_buffer = post_frame
            return _core.GateCloseFrame(pre_frame)

    def _collect_post_trigger_samples(self, frame):
        if self._remaining_post_trigger_samples > frame.frame.shape[1]:
            self._remaining_post_trigger_samples -= frame.frame.shape[1]
            return frame

        self.open = False
        self.waiting = True  # TODO: Always reactivate?
        pre_frame, post_frame = frame.frame[:, :self._remaining_post_trigger_samples], frame.frame[:, self._remaining_post_trigger_samples:]
        frame = _core.GateCloseFrame(pre_frame)
        self._remaining_post_trigger_samples = 0
        self._post_close_buffer = post_frame
        return frame


class Trigger(_core.SamplerateFollower):
    def __init__(self, channel=0, **kwargs):
        super().__init__(**kwargs)
        self.channel = channel

    def to_dict(self):
        return super().to_dict() | dict(
            channel=self.channel,
        )

    def process(self, frame):
        if isinstance(frame, _core.Frame):
            indices = self.detect(frame.frame[self.channel])
            if indices is not None:
                frame = _core.TriggerFrame(frame.frame, indices)
        return frame

    def detect(self, frame):
        ...


class PeakTrigger(Trigger):
    def __init__(self, threshold=0, edge='rising', **kwargs):
        super().__init__(**kwargs)
        self.threshold = threshold
        self.edge = edge
        self._prev_val = None

    def to_dict(self):
        return super().to_dict() | dict(
            threshold=self.threshold,
            edge=self.edge,
        )

    def reset(self, **kwargs):
        super().reset(**kwargs)
        self._prev_val = None

    @property
    def threshold(self):
        return self._threshold

    @threshold.setter
    def threshold(self, val):
        self._threshold = val

    def detect(self, frame):
        if self._prev_val is None:
            self._prev_val = frame[0]
        signs = np.sign(np.insert(frame, 0, self._prev_val) - self._threshold)
        self._prev_val = frame[-1]

        if self.edge == 'crossing':
            test = np.diff(signs)
        elif self.edge == 'rising':
            test = np.diff(signs) > 0
        elif self.edge == 'falling':
            test = np.diff(signs) < 0
        elif self.edge == 'above':
            test = signs > 0
        elif self.edge == 'below':
            test = signs < 0
        if np.any(test):
            # Activation met, we'll stop tracking the levels!
            return np.nonzero(test)[0]


class RMSTrigger(PeakTrigger):
    def __init__(self, threshold, time_constant, **kwargs):
        super().__init__(threshold=threshold, **kwargs)
        self.time_constant = time_constant
        self._state = None

    def to_dict(self):
        return super().to_dict() | dict(
            time_constant=self.time_constant,
        )

    @property
    def threshold(self):
        return self._threshold ** 0.5

    @threshold.setter
    def threshold(self, val):
        self._threshold = val ** 2

    def setup(self, **kwargs):
        super().setup(**kwargs)
        self._digital_time_coeff = np.exp(-1 / (self.time_constant * self.samplerate))

    def reset(self, **kwargs):
        super().reset(**kwargs)
        self._state = None

    def detect(self, frame):
        if self._state is None:
            self._state = scipy.signal.lfilter_zi(
                [1 - self._digital_time_coeff],
                [1, -self._digital_time_coeff],
            ) * frame[0]**2
        levels, self._state = scipy.signal.lfilter(
            [1 - self._digital_time_coeff],
            [1, -self._digital_time_coeff],
            frame**2, zi=self._state
        )

        return super().detect(levels)



from . import _core, signal_tools
import numpy as np
import scipy.signal


class _Generator(_core.SamplerateFollower):
    def process(self, packet):
        if not isinstance(packet, _core.FrameRequest):
            return packet
        frame = self.generate(packet.framesize)
        if isinstance(frame, _core.Frame):
            frame.frame = np.atleast_2d(frame.frame)
        elif not isinstance(frame, _core.Packet):
            frame = _core.Frame(np.atleast_2d(frame))
        return frame

    def generate(self, framesize):
        return np.zeros(framesize)


class SignalGenerator(_Generator):
    def __init__(
        self, signal=None, repetitions=1,
        amplitude=1,
        fade_in=None, fade_out=None,
        pre_pad=None, post_pad=None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.amplitude = amplitude
        self.fade_in = fade_in
        self.fade_out = fade_out
        self.pre_pad = pre_pad
        self.post_pad = post_pad

        if signal is not None:
            self.signal = signal

        if repetitions == -1 or (isinstance(repetitions, str) and repetitions.lower()[:3] == 'inf'):
            repetitions = np.inf
        self.repetitions = repetitions
        self.reset()

    def to_dict(self):
        return super().to_dict() | dict(
            amplitude=self.amplitude,
            fade_in=self.fade_in,
            fade_out=self.fade_out,
            pre_pad=self.pre_pad,
            post_pad=self.post_pad,
            repetitions=self.repetitions,
        )

    def reset(self, **kwargs):
        super().reset(**kwargs)
        self._repetitions_done = 0
        self._sample_index = 0

    @property
    def signal(self):
        try:
            return self._signal
        except AttributeError:
            pass
        self.setup()
        return self._signal

    @signal.setter
    def signal(self, signal):
        signal = signal_tools.fade_signals(signal, fade_in=self.fade_in, fade_out=self.fade_out, samplerate=self.samplerate, inplace=False)
        signal = signal_tools.pad_signals(signal, pre_pad=self.pre_pad, post_pad=self.post_pad, samplerate=self.samplerate)
        self._signal = signal * self.amplitude

    def generate(self, framesize):
        if self._repetitions_done >= self.repetitions:
            return _core.LastFrame(np.zeros(framesize), 0)

        start_idx = self._sample_index
        stop_idx = self._sample_index + framesize
        signal_length = self.signal.shape[-1]

        if stop_idx <= signal_length:
            frame = self.signal[..., start_idx:stop_idx]
        else:
            num_repeats = np.math.ceil(stop_idx / signal_length)
            repeated_signal = np.tile(self.signal, num_repeats)
            frame = repeated_signal[..., start_idx:stop_idx]

        repeats, self._sample_index = divmod(stop_idx, signal_length)
        self._repetitions_done += repeats
        if self._repetitions_done >= self.repetitions:
            if self._sample_index > 0:
                frame[..., -self._sample_index:] = 0
            frame = _core.LastFrame(frame, framesize - self._sample_index)
        else:
            frame = _core.Frame(frame)
        return frame

    def once(self):
        if self._sample_index != 0 and self._repetitions_done != 0:
            raise RuntimeError('Cannot use method `once` on generator which is running in a streamed pipeline!')
        framesize = self.repetitions * self.signal.shape[-1]

        frame = self.request(_core.FrameRequest(framesize))
        self._sample_index = self._repetitions_done = 0
        return self.push(frame)


class SweepGenerator(SignalGenerator):
    def __init__(
        self,
        lower_frequency, upper_frequency,
        sweep_length, method='logarithmic',
        amplitude_slope=None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.lower_frequency = lower_frequency
        self.upper_frequency = upper_frequency
        self.sweep_length = sweep_length
        self.amplitude_slope = amplitude_slope
        self.method = method

    def to_dict(self):
        return super().to_dict() | dict(
            lower_frequency=self.lower_frequency,
            upper_frequency=self.upper_frequency,
            sweep_length=self.sweep_length,
            amplitude_slope=self.amplitude_slope,
            method=self.method,
        )

    def setup(self, **kwargs):
        super().setup(**kwargs)
        time_vector = np.arange(round(self.sweep_length * self.samplerate)) / self.samplerate
        signal = scipy.signal.chirp(t=time_vector, f0=self.lower_frequency, t1=self.sweep_length, f1=self.upper_frequency, method=self.method, phi=-90)
        last_crossing = -1
        while np.sign(signal[last_crossing - 1]) == np.sign(signal[-1]):
            last_crossing -= 1
        signal[last_crossing:] = 0
        if self.amplitude_slope is not None:
            end_value = 10**(6 * self.amplitude_slope / 20 * np.log2(self.upper_frequency / self.lower_frequency))
            if self.method == 'logarithmic':
                slope = np.geomspace(1, end_value, signal.size)
            elif self.method == 'linear':
                slope = np.linspace(1, end_value, signal.size)
            else:
                raise ValueError(f'Cannot use amplitude compensation with sweep method {self.method}')
            slope /= max(slope)
        else:
            slope = 1
        self.reference = signal.copy()
        self.signal = signal * slope


class MaximumLengthGenerator(SignalGenerator):
    def __init__(self, order=None, sequence_time=None, **kwargs):
        super().__init__(**kwargs)
        if None not in (order, sequence_time):
            raise ValueError('Cannot give both MLS order and MLS time')
        elif sequence_time is not None:
            self._sequence_time = sequence_time
        elif order is not None:
            self.order = order

    def to_dict(self):
        return super().to_dict() | dict(
            order=self.order,
        )

    @property
    def sequence_time(self):
        try:
            return self._sequence_time
        except AttributeError:
            pass
        try:
            return (2**self.order - 1) / self.samplerate
        except AttributeError:
            raise AttributeError('Cannot use MLS sequence time without a defined samplerate')

    @sequence_time.setter
    def sequence_time(self, value):
        try:
            n_samples = self.samplerate * value
        except AttributeError:
            self._sequence_time = value
            del self._order
        else:
            self.order = np.math.ceil(np.log2(n_samples + 1))

    @property
    def order(self):
        try:
            return self._order
        except AttributeError:
            pass
        try:
            n_samples = self.samplerate * self._sequence_time
        except AttributeError:
            raise AttributeError('Cannot use MLS sequence time without a defined samplerate')
        return np.math.ceil(np.log2(n_samples + 1))

    @order.setter
    def order(self, value):
        if not isinstance(value, int):
            raise TypeError(f"Cannot set order to non-integer value {value}")
        self._order = value
        try:
            del self._sequence_time
        except AttributeError:
            pass

    def setup(self, **kwargs):
        super().setup(**kwargs)
        self.reference, _ = scipy.signal.max_len_seq(self.order)
        self.signal = (1 - 2 * self.reference).astype('float64')


class ToneGenerator(_Generator):
    _functions = {
        'sin': np.sin,
        'saw': scipy.signal.sawtooth,
        'squ': scipy.signal.square
    }

    def __init__(
        self,
        frequency,
        amplitude=1,
        periods=None, duration=None,
        phase_offset=0,
        shape='sine',
        shape_kwargs=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.frequency = frequency
        self.amplitude = amplitude
        self.shape = shape
        self.shape_kwargs = shape_kwargs if shape_kwargs is not None else {}
        self.phase_offset = phase_offset

        if periods is None and duration is None:
            self.periods = np.inf
        elif periods is not None:
            self.periods = periods
        elif duration is not None:
            self.periods = duration * self.frequency
        else:
            raise ValueError('Cannot specify both duration and number of repetitions')

        self.reset()

    def to_dict(self):
        return super().to_dict() | dict(
            frequency=self.frequency,
            shape=self.shape,
            phase_offset=self.phase_offset,
            periods=self.periods,
            **self.shape_kwargs,
        )

    def reset(self, **kwargs):
        super().reset(**kwargs)
        self._phase = 0

    def generate(self, framesize):
        if self._phase >= self.periods * 2 * np.pi:
            return _core.LastFrame(np.zeros(framesize), 0)

        phase = np.arange(framesize) * (2 * np.pi * self.frequency / self.samplerate)
        signal = self._function(self._phase + self.phase_offset + phase, **self.shape_kwargs) * self.amplitude
        self._phase += 2 * np.pi * framesize * self.frequency / self.samplerate
        if self._phase > self.periods * 2 * np.pi:
            phase_to_mute = self._phase - self.periods * 2 * np.pi
            samples_to_mute = np.math.floor(phase_to_mute / (2 * np.pi * self.frequency) * self.samplerate)
            signal[-samples_to_mute:] = 0
            frame = _core.LastFrame(signal, framesize - samples_to_mute)
        else:
            frame = _core.Frame(signal)
        return frame

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, value):
        if value[:3].lower() in self._functions:
            self._shape = value
            self._function = self._functions[value[:3].lower()]
        else:
            raise ValueError(f'Unknown tone shape {value}')



import datetime
import re
import os

# TODO: make a timestamper class instead of functions?


def parse_timestamp(ts):
    sep = r'[-,_,:,T, ]?'

    year = r'(?P<year>\d{4})'
    month = r'(?P<month>\d{2})'
    day = r'(?P<day>\d{2})'
    hour = r'(?P<hour>\d{2})?'
    minute = r'(?P<minute>\d{2})?'
    second = r'(?P<second>\d{2})?'
    fraction = r'\.?(?P<fraction>\d+)?'

    pattern = sep.join([year, month, day, hour, minute, second]) + fraction
    regex = re.compile(pattern)
    match = regex.search(ts)
    if match is None:
        raise ValueError(f'Count not parse timestamp {ts}')

    #     raise ValueError(f'Count not parse timestamp {ts}')

    args = []
    for arg in match.groups():
        if arg is not None:
            args.append(int(arg))
        else:
            break
    return datetime.datetime(*args)
    # TODO: convert the matched strings to ints
    # TODO: fill two digit year to four digit year
    # TODO: raise something meaningful on missing info
    return datetime.datetime(
        int(groups['year']),
        int(groups['month']),
        int(groups['day']),
        int(groups['hour']),
        int(groups['minute']),
        int(groups['second']),
        int(groups['fraction']),
    )


def timestamp(format='ISO', microseconds=False, utc=False):
    now = datetime.datetime.now(datetime.timezone.utc)
    if not utc:
        now = now.astimezone()
    if format.lower() == 'iso':
        return now.isoformat(timespec='microseconds' if microseconds else 'seconds')
    if format.lower() == 'filename':
        return now.strftime('%Y-%m-%d_%H-%M-%S' + ('.%f' if microseconds else ''))
    if format.lower() == 'epoch':
        return now.timestamp() if microseconds else int(now.timestamp())
    raise ValueError(f"Unknown timestamp format {format}")


def newest_timestamed_filename(directory, n_files=1):
    return_files = []
    files = os.listdir(directory)
    for _ in range(n_files):
        newest = None
        for file in files:
            try:
                stamp = parse_timestamp(file)
            except ValueError:
                continue
            if newest is None:
                newest = (file, stamp)
            elif stamp > newest[1]:
                newest = (file, stamp)
        if newest is not None:
            return_files.append(newest[0])
            files.remove(file)
    if len(return_files) < n_files:
        raise ValueError('No matching files found')
    if n_files == 1:
        return return_files[0]
    return return_files



import logging
logger = logging.getLogger(__name__)

from . import _version  # noqa: E402
__version_info__ = _version.version_info
__version__ = _version.version
del _version

from . import interfaces, generators, recorders, analysis, signal_tools, timestamps, triggers  # noqa: E402, F401



"""File generated while packaging."""
import contextlib

version_info = {'short': '0.2.1', 'full': '0.2.1', 'release': True, 'clean': True, 'git revision': 'a4b6d5f80589840eadf8a5a9b9908565ba06bb57'}
version = '0.2.1'


@contextlib.contextmanager
def version_manager():
    """Dummy context manager, returns the version."""
    yield version



from setuptools import setup
import sys
import os.path


sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'acoustics_hardware'))
from _version import version_manager  # We cannot import the _version module, but we can import from it.

with version_manager() as version:
    setup(
        name='acoustics-hardware',
        version=version,
        description='Controlling hardware used in acoustic measurement systems',
        long_description=open('README.rst').read(),
        long_description_content_type='text/x-rst',
        url='https://github.com/AppliedAcousticsChalmers/acoustics-hardware',
        author='Carl Andersson',
        author_email='carl.andersson@chalmers.se',
        license='MIT',
        packages=['acoustics_hardware'],
        install_requires=[
            'numpy',
            'scipy',
            'sounddevice',
            'nidaqmx;platform_system=="Windows"',
            'pyserial',
            'SchunkMotionProtocol',
            'zarr',
        ],
    )



from . import _core, signal_tools
import numpy as np
import sounddevice as sd
import time
import functools
import warnings

try:
    import nidaqmx
    import nidaqmx.stream_readers
    import nidaqmx.stream_writers
except ImportError:
    pass


def _channel_collection(channel_class, **default_channel_kwargs):
    @functools.wraps(channel_class, updated=())
    class ChannelCollection:
        default_channel_settings = default_channel_kwargs
        def __init__(self, channels=None):
            self.channels = []
            if channels is None:
                return

            if isinstance(channels, int):
                # Channels given as number of channels to use
                channels = range(channels)
            if isinstance(channels, dict):
                # Channels given as a single dict defining one channel
                channels = [channels]

            try:
                # Channels given as a single pair of (int, dict) for a single channel
                ch, configs = channels
            except (ValueError, TypeError):
                pass
            else:
                if isinstance(ch, int) and isinstance(configs, dict):
                    # This extra check is needed in case channels is given as e.g. [0, 1]
                    # which defines two channels, not one.
                    channels = [channels]

            for channel in channels:
                try:
                    # This channel is a single dict defining the channel
                    self.append(**channel)
                except TypeError:
                    pass
                else:
                    continue

                try:
                    # Channel is an (int, dict) pair
                    ch, configs = channel
                except (ValueError, TypeError):
                    pass
                else:
                    if isinstance(ch, int) and isinstance(configs, dict):
                        self.append(channel=ch, **configs)
                        continue

                if isinstance(channel, int):
                    # This channel i s just defined by its index, with no settings
                    self.append(channel)
                    continue
                raise ValueError(f'Cannot add channel with definition {channel}')


        def append(self, channel, **kwargs):
            self.channels.append(channel_class(channel, **self.default_channel_settings | kwargs))

        def __getitem__(self, key):
            return self.channels[key]

        def __len__(self):
            return len(self.channels)

        def __repr__(self):
            return repr(self.channels)

    return ChannelCollection


class SimpleChannel:
    """A single channel without settings.

    Parameters
    -----------
    channel : int
        The index of the channel, zero based.
    label : string, default none
        An optional label to assign to the channel.
    """

    def __init__(self, channel, label=None):
        self.channel = channel
        self.label = label

    def __repr__(self):
        label = '' if self.label is None else f", '{self.label}'"
        return f'{self.__class__.__name__}({self.channel}{label})'


class _StreamedInterface(_core.SamplerateDecider):
    _input_channels = _channel_collection(SimpleChannel)
    _output_channels = _channel_collection(SimpleChannel)

    def __init__(self, input_channels=None, output_channels=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.input_channels = self._input_channels(input_channels)
        self.output_channels = self._output_channels(output_channels)

    def to_dict(self):
        return super().to_dict() | dict(
            input_channels=repr(self.input_channels),
            output_channels=repr(self.output_channels),
        )

    def process(self, packet):
        frame = framesize = None
        if isinstance(packet, _core.Frame):
            frame = packet.frame
        elif isinstance(packet, _core.FrameRequest):
            framesize = packet.framesize
        elif not isinstance(packet, _core.Packet):
            if np.ndim(packet) == 0:
                framesize = packet
            else:
                frame = packet
        return _core.Frame(self.playrec(frame=frame, framesize=framesize))


class DummyInterface(_StreamedInterface):
    def __init__(self, framesize=None, maxframes=np.inf, realtime=False, threaded=True, **kwargs):
        super().__init__(**kwargs)
        self.framesize = framesize
        self.maxframes = maxframes
        self.realtime = realtime
        self.threaded = threaded

    def run(self):
        if self.threaded:
            import threading
            self._running = threading.Event()
            self._thread = threading.Thread(target=self._passthrough_target)
            self._thread.start()
        else:
            frames = 0
            while frames < self.maxframes:
                frames += 1
                frame = self._upstream.request(_core.FrameRequest(self.framesize))
                if self.realtime:
                    time.sleep(self.framesize / self.samplerate)
                self._downstream.push(frame)
                if isinstance(frame, _core.LastFrame):
                    break

    def _passthrough_target(self):
        self._running.set()
        frames = 0
        while self._running.is_set() and frames < self.maxframes:
            frames += 1
            frame = self._upstream.request(_core.FrameRequest(self.framesize))
            if self.realtime:
                time.sleep(self.framesize / self.samplerate)
            self._downstream.push(frame)
            if isinstance(frame, _core.LastFrame):
                self._running.clear()
                break

    def playrec(self, frame=None, framesize=None):
        if frame is not None:
            return frame
        return np.zeros((1, framesize))

    def stop(self):
        self._running.clear()


class AudioInterface(_StreamedInterface):
    @classmethod
    def list_interfaces(cls):
        """Check which audio interfaces can be interacted with.

        This is just a wrapper around `sounddevice.query_devices`.

        Returns
        -------
        interfaces : list
            A list of all interfaces found.
        """
        return sd.query_devices()

    def __init__(self, name=None, input_name=None, output_name=None, framesize=None, **kwargs):
        super().__init__(**kwargs)
        self.framesize = framesize
        input_name = input_name or name
        output_name = output_name or name
        input_name = sd.default.device[0] if input_name is None else input_name
        output_name = sd.default.device[1] if output_name is None else output_name

        device_list = sd.query_devices()
        self._input_device = sd.query_devices(input_name)
        self._input_device['index'] = device_list.index(self._input_device)
        self._output_device = sd.query_devices(output_name)
        self._output_device['index'] = device_list.index(self._output_device)

    def to_dict(self):
        return super().to_dict() | dict(
            input_device=self._input_device['name'],
            output_device=self._output_device['name'],
            framesize=self.framesize,
        )

    def run(self):
        outputs = [ch.channel for ch in self.output_channels]
        inputs = [ch.channel for ch in self.input_channels]
        num_input_ch = max(inputs, default=-1) + 1
        num_output_ch = max(outputs, default=-1) + 1
        if num_output_ch == 1 and self.max_outputs > 1:
            # PortAudio by default duplicates mono channels over all channels.
            # This is not what the rest of this package is doing.
            num_output_ch = 2
        silent_ch = [ch for ch in range(num_output_ch) if ch not in outputs]

        def process_input(frame):
            # Take the frame read from hardware and push it downstream in the pipeline.
            frame = _core.Frame(frame.T[inputs])
            self._downstream.push(frame)

        def process_output(buffer):
            # Get a frame from upstream in the pipeline and write it to the buffer.
            frame = self._upstream.request(_core.FrameRequest(buffer.shape[0]))
            buffer[:, outputs] = frame.frame.T
            buffer[:, silent_ch] = 0
            if isinstance(frame, _core.LastFrame):
                raise sd.CallbackStop()

        if num_input_ch and num_output_ch:
            def callback(indata, outdata, frames, time, status):
                process_input(indata)
                process_output(outdata)
            self._stream = sd.Stream(
                samplerate=self.samplerate,
                blocksize=self.framesize,
                device=(self._input_device['index'], self._output_device['index']),
                channels=(num_input_ch, num_output_ch),
                callback=callback,
            )
        elif num_input_ch:
            def callback(indata, frames, time, status):
                process_input(indata)
            self._stream = sd.InputStream(
                samplerate=self.samplerate,
                blocksize=self.framesize,
                device=self._input_device['name'],
                channels=num_input_ch,
                callback=callback,
            )
        elif num_output_ch:
            def callback(outdata, frames, time, status):
                process_output(outdata)
            self._stream = sd.OutputStream(
                samplerate=self.samplerate,
                blocksize=self.framesize,
                device=self._output_device['index'],
                channels=num_output_ch,
                callback=callback,
            )

        self._stream.start()

    def stop(self):
        self._stream.stop()

    def playrec(self, frame=None, framesize=None):
        outputs = [ch.channel + 1 for ch in self.output_channels]
        inputs = [ch.channel + 1 for ch in self.input_channels]

        if len(inputs) and len(outputs):
            return sd.playrec(
                data=frame.T,
                samplerate=self.samplerate,
                device=(self._input_device['index'], self._output_device['index']),
                input_mapping=inputs,
                output_mapping=outputs,
                blocking=True,
            )
        elif len(inputs):
            return sd.rec(
                frames=framesize,
                samplerate=self.samplerate,
                device=self._input_device['index'],
                mapping=inputs,
                blocking=True,
            )
        elif len(outputs):
            return sd.play(
                data=frame.T,
                samplerate=self.samplerate,
                device=self._output_device['index'],
                mapping=outputs,
                blocking=True,
            )

    @property
    def max_inputs(self):
        return self._input_device['max_input_channels']

    @property
    def max_outputs(self):
        return self._output_device['max_output_channels']


class NationalInstrumentsConfigurableChannel(SimpleChannel):
    """A configurable National instruments channel

    Parameters
    ----------
    channel : int
        The index of the channel, zero based.
    label : str, default None
        Optional label to assign the channel
    voltage_range : numeric
        The range in which to configure the channel.
        Give a two values (min, max) or a single amplitude value.
    terminal_config : str, default None
        How to configure the terminals for the channel. Leave as None to use device default configuration.
        Should be one of
        - Differential: for a differential input
        - Single ended: for a single ended measurement
    """
    def __init__(self, channel, label=None, voltage_range=None, terminal_config=None):
        super().__init__(channel=channel, label=label)
        self.voltage_range = voltage_range
        self.terminal_config = terminal_config

    def config(self, interface):
        configs = {}
        if self.voltage_range is not None:
            try:
                min_val, max_val = self.voltage_range
            except TypeError:
                min_val = -self.voltage_range
                max_val = self.voltage_range
            configs['min_val'] = min_val
            configs['max_val'] = max_val
        if self.terminal_config is not None:
            if 'diff' in self.terminal_config.lower():
                configs['terminal_config'] = nidaqmx.constants.TerminalConfiguration.DIFF
            elif 'single' in self.terminal_config.lower():
                configs['terminal_config'] = nidaqmx.constants.TerminalConfiguration.RSE
            else:
                raise ValueError(f'Unknown terminal configuration {self.terminal_config}')
        return configs

    def __repr__(self):
        s = super().__repr__()[:-1]
        voltage_range = '' if self.voltage_range is None else f', voltage_range={self.voltage_range}'
        terminal_config = '' if self.terminal_config is None else f', terminal_config={self.terminal_config}'
        return s + voltage_range + terminal_config + ')'


class NationalInstrumentsInputChannel(NationalInstrumentsConfigurableChannel):
    def config(self, interface):
        return super().config(interface) | dict(physical_channel=f'{interface.name}/ai{self.channel}')


class NationalInstrumentsOutputChannel(NationalInstrumentsConfigurableChannel):
    def config(self, interface):
        return super().config(interface) | dict(physical_channel=f'{interface.name}/ao{self.channel}')


class NationalInstrumentsDaqmx(_StreamedInterface):
    @classmethod
    def list_interfaces(cls):
        """Check what National Instruments hardware is available.

        Returns
        -------
        interfaces : list
            A list of all devices.
        """
        try:
            system = nidaqmx.system.System.local()
        except NameError as e:
            if e.args[0] == "name 'nidaqmx' is not defined":
                raise ModuleNotFoundError("Windows-only module 'nidaqmx' is not installed")
            else:
                raise e
        name_list = [dev.name for dev in system.devices]
        return name_list

    def __init__(self, name=None, framesize=None, buffer_n_frames=25, **kwargs):
        self.name = name or self.list_interfaces()[0]
        self._device = nidaqmx.system.Device(self.name)
        self._input_channels = _channel_collection(
            NationalInstrumentsInputChannel,
            terminal_config='single ended',
            voltage_range=(min(self._device.ai_voltage_rngs), max(self._device.ai_voltage_rngs))
        )
        self._output_channels = _channel_collection(
            NationalInstrumentsOutputChannel,
            voltage_range=(min(self._device.ao_voltage_rngs), max(self._device.ao_voltage_rngs))
        )
        super().__init__(**kwargs)
        self.framesize = framesize
        self.buffer_n_frames = buffer_n_frames

    def to_dict(self):
        return super().to_dict() | dict(
            name=self.name,
            framesize=self.framesize,
        )

    @property
    def samplerate(self):
        return self._samplerate

    @samplerate.setter
    def samplerate(self, val):
        has_in = self.max_inputs > 0
        has_out = self.max_outputs > 0

        if val is None:
            if has_in and has_out:
                fs_in = self._device.ai_max_multi_chan_rate
                fs_out = self._device.ao_max_rate
            elif has_in:
                fs_in = fs_out = self._device.ai_max_multi_chan_rate
            elif has_out:
                fs_in = fs_out = self._device.ao_max_rate
            else:
                raise ValueError(f'National Instruments interface {self.name} seems to have neither inputs nor outputs')
        else:
            try:
                fs_out, fs_in = val
            except TypeError:
                fs_out = fs_in = val

        if has_in:
            task = nidaqmx.Task()
            task.ai_channels.add_ai_voltage_chan(self._device.ai_physical_chans[0].name)
            task.timing.cfg_samp_clk_timing(rate=fs_in)
            if not fs_in == task.timing.samp_clk_rate:
                warnings.warn(f'Target input samplerate {fs_in} not available on device, using {task.timing.samp_clk_rate}', stacklevel=2)
                fs_in = task.timing.samp_clk_rate
            task.close()

        if has_out:
            task = nidaqmx.Task()
            task.ao_channels.add_ao_voltage_chan(self._device.ao_physical_chans[0].name)
            task.timing.cfg_samp_clk_timing(rate=fs_out)
            if not fs_out == task.timing.samp_clk_rate:
                warnings.warn(f'Target output samplerate {fs_out} not available on device, using {task.timing.samp_clk_rate}', stacklevel=3)
                fs_out = task.timing.samp_clk_rate
            task.close()

        if fs_out == fs_in:
            self._samplerate = fs_in
        else:
            self._samplerate = (fs_out, fs_in)

    @property
    def framesize(self):
        return self._framesize

    @framesize.setter
    def framesize(self, val):
        if val is None:
            fs = self.samplerate
            try:
                fs_out, fs_in = fs
            except TypeError:
                fs_in = fs_out = fs
            N_in = fs_in // 50
            N_out = fs_out // 50
        else:
            try:
                N_out, N_in = val
            except TypeError:
                N_out = N_in = val

        if N_out == N_in:
            self._framesize = N_out
        else:
            self._framesize = (N_out, N_in)

    def reset(self, **kwargs):
        super().reset(**kwargs)
        try:
            self._device.reset_device()
        except (AttributeError, nidaqmx.DaqError):
            pass

    @property
    def max_inputs(self):
        return len(self._device.ai_physical_chans)

    @property
    def max_outputs(self):
        return len(self._device.ao_physical_chans)

    def setup(self, finite_samples=False, **kwargs):
        super().setup(**kwargs)

        if len(self.input_channels):
            self._input_task = nidaqmx.Task()
            for ch in self.input_channels:
                self._input_task.ai_channels.add_ai_voltage_chan(**ch.config(self))

        if len(self.output_channels):
            self._output_task = nidaqmx.Task()
            for ch in self.output_channels:
                self._output_task.ao_channels.add_ao_voltage_chan(**ch.config(self))

        try:
            samplerate_upstream, samplerate_downstream = self.samplerate
        except TypeError:
            samplerate_upstream = samplerate_downstream = self.samplerate

        if finite_samples is False:
            try:
                framesize_upstream, framesize_downstream = self.framesize
            except TypeError:
                framesize_upstream = framesize_downstream = self.framesize
            samps_per_chan_read = self.buffer_n_frames * framesize_downstream
            samps_per_chan_write = self.buffer_n_frames * framesize_upstream
            sample_mode = nidaqmx.constants.AcquisitionType.CONTINUOUS
        else:
            try:
                samps_per_chan_write, samps_per_chan_read = finite_samples
            except TypeError:
                samps_per_chan_write = samps_per_chan_read = finite_samples
            sample_mode = nidaqmx.constants.AcquisitionType.FINITE

        if len(self.input_channels):
            self._input_task.timing.cfg_samp_clk_timing(
                rate=samplerate_downstream,
                samps_per_chan=samps_per_chan_read,
                sample_mode=sample_mode
            )
            if finite_samples is False:
                self._databuffer = np.empty((self._input_task.in_stream.num_chans, framesize_downstream))
                self._read = nidaqmx.stream_readers.AnalogMultiChannelReader(self._input_task.in_stream).read_many_sample
                self._samples_read = 0
                self._input_task.in_stream.input_buf_size = framesize_downstream * self.buffer_n_frames
                self._input_task.register_every_n_samples_acquired_into_buffer_event(framesize_downstream, self._read_callback)

        if len(self.output_channels):
            self._output_task.timing.cfg_samp_clk_timing(
                rate=samplerate_upstream,
                samps_per_chan=samps_per_chan_write,
                sample_mode=sample_mode
            )

            if finite_samples is False:
                self._write = nidaqmx.stream_writers.AnalogMultiChannelWriter(self._output_task.out_stream).write_many_sample
                self._output_task.out_stream.regen_mode = nidaqmx.constants.RegenerationMode.DONT_ALLOW_REGENERATION
                self._samples_written = 0
                self._output_task.out_stream.output_buf_size = framesize_upstream * self.buffer_n_frames
                self._write_callback(None, None, self.buffer_n_frames * framesize_upstream, None)
                self._output_task.register_every_n_samples_transferred_from_buffer_event(framesize_upstream, self._write_callback)

        if len(self.output_channels) and len(self.input_channels):
            self._input_task.triggers.start_trigger.cfg_dig_edge_start_trig(f'/{self.name}/ao/StartTrigger')


    def run(self):
        if not self._is_ready:
            self.setup()

        if len(self.input_channels):
            self._input_task.start()
        if len(self.output_channels):
            self._output_task.start()

    def stop(self):
        self._is_ready = False  # The tasks are single use!

        if len(self.output_channels):
            try:
                self._output_task.stop()
                self._output_task.wait_until_done(timeout=10)
                self._output_task.close()
            except nidaqmx.DaqError:
                pass

        if len(self.input_channels):
            if len(self.output_channels):
                while self._samples_read < self._samples_written:
                    time.sleep(self._input_task.in_stream.sleep_time)
            try:
                self._input_task.stop()
                self._input_task.wait_until_done(timeout=10)
                self._input_task.close()
            except nidaqmx.DaqError:
                pass

    def _read_callback(self, task_handle, every_n_samples_event_type, number_of_samples, callback_data):
        try:
            self._read(self._databuffer, number_of_samples)
        except nidaqmx.DaqError:
            self.stop()
            raise
        self._samples_read += number_of_samples
        self._downstream.push(_core.Frame(self._databuffer.copy()))
        return 0

    def _write_callback(self, task_handle, every_n_samples_event_type, number_of_samples, callback_data):
        frame = self._upstream.request(_core.FrameRequest(number_of_samples))
        self._write(frame.frame)
        self._samples_written += number_of_samples
        if isinstance(frame, _core.LastFrame):
            self._samples_written -= number_of_samples
            self._samples_written += frame.valid_samples
            frame = self._write(np.zeros((len(self.output_channels), number_of_samples)))
            while self._output_task.out_stream.total_samp_per_chan_generated < self._samples_written:
                time.sleep(self._output_task.out_stream.sleep_time)
            self.stop()
        return 0

    def playrec(self, frame=None, framesize=None):
        if frame is not None:
            padded_frame = signal_tools.pad_signals(frame, post_pad=1)
        if framesize is None and len(self.input_channels):
            try:
                samplerate_upstream, samplerate_downstream = self.samplerate
            except TypeError:
                samplerate_upstream = samplerate_downstream = self.samplerate
            framesize = np.math.ceil(padded_frame.shape[-1] * samplerate_downstream / samplerate_upstream)
        self.setup(finite_samples=(padded_frame.shape[-1], framesize), pipeline=True)

        if len(self.input_channels):
            self._input_task.start()

        if len(self.output_channels):
            timeout = frame.shape[-1] / samplerate_upstream * 1.2 + 5
            self._output_task.write(padded_frame.squeeze())
            self._output_task.start()
            self._output_task.wait_until_done(timeout=timeout)
            self._output_task.close()

        if len(self.input_channels):
            timeout = framesize / samplerate_downstream * 1.2 + 5
            read_frame = self._input_task.read(framesize, timeout=timeout)
            read_frame = np.asarray(read_frame)
            self._input_task.close()
            return read_frame[..., 1:]


class NationalInstrumentsMultimoduleInputChannel(NationalInstrumentsConfigurableChannel):
    def __init__(self, module, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.module = module

    def __repr__(self):
        s = super().__repr__()
        pre, post = s.split('(')
        return f'{pre}(module={self.module}, {post}'

    def config(self, interface):
        return interface.input_modules[self.module].ai_physical_chans[self.channel].name


class NationalInstrumentsMultimoduleOutputChannel(NationalInstrumentsMultimoduleInputChannel):
    def config(self, interface):
        return interface.output_modules[self.module].ao_physical_chans[self.channel].name


class CompactDaqmxMultimodule(NationalInstrumentsDaqmx):
    @classmethod
    def get_chassis(cls, name=None):
        devices = cls.list_interfaces()
        chassis = [dev for dev in devices if 'cDAQ' in dev and 'Mod' not in dev]
        if name is None:
            return chassis
        if name not in chassis:
            name = [dev for dev in chassis if str(name) in dev][0]
        return nidaqmx.system.Device(name)

    @classmethod
    def get_modules(cls, chassis):
        try:
            chassis = chassis.name
        except AttributeError:
            pass
        chassis = cls.get_chassis(chassis)
        devs = cls.list_interfaces()
        return [nidaqmx.system.Device(dev) for dev in devs if chassis in dev and 'Mod' in dev]

    @classmethod
    def reset_chassis_modules(cls, chassis):
        for module in cls.get_modules(chassis):
            nidaqmx.system.Device(module).reset_device()

    _input_channels = _channel_collection(NationalInstrumentsMultimoduleInputChannel)
    _output_channels = _channel_collection(NationalInstrumentsMultimoduleOutputChannel)

    def __init__(self, name=None, **kwargs):
        if name is None:
            name = self.list_devices()[0]
        self.chassis = self.get_chassis(name)
        self.modules = self.get_modules(self.chassis)

        # We need to set a default name for the NIDAQ class to initialize properly
        if len(self.input_modules) > 0:
            self.name = self.input_modules[0]
        elif len(self.output_modules) > 0:
            self.name = self.output_modules[0]
        super().__init__(**kwargs)
        del self.name

        self.input_modules = [module for module in self.modules if len(module.ai_physical_chans) > 0]
        self.output_modules = [module for module in self.modules if len(module.ao_physical_chans) > 0]

    @property
    def max_inputs(self):
        inputs = 0
        for module in self.input_modules:
            inputs += len(module.ai_physical_chans)
        return inputs

    @property
    def max_outputs(self):
        outputs = 0
        for module in self.output_modules:
            outputs += len(module.ao_physical_chans)
        return outputs

    def setup(self, **kwargs):
        super().setup(**kwargs)
        if len(self.input_channels) and len(self.output_channels):
            self._output_task.timing.samp_clk_timebase_src = self._input_task.timing.samp_clk_timebase_src
            self._output_task.timing.samp_clk_timebase_rate = self._input_task.timing.samp_clk_timebase_rate

    def reset(self, **kwargs):
        super().reset(**kwargs)
        for module in self.input_modules:
            module.reset_device()
        for module in self.output_modules:
            module.reset_device()



import numpy as np
import scipy.signal
from . import signal_tools


def dB(x, power=False, safe_zeros=True):
    if safe_zeros and np.size(x) > 1:
        x = signal_tools.nonzero_signals(x)
    if power:
        if np.any(x < 0):
            if power == 'imag':
                return 10 * np.log10(x + 0j)
            if power == 'nan':
                return 10 * np.log10(np.where(x > 0, x, np.nan))
        return 10 * np.log10(x)
    else:
        return 20 * np.log10(np.abs(x))


def sweep_deconvolution(
    reference_signal,
    measured_signal,
    samplerate=None,
    lower_cutoff=None,
    upper_cutoff=None,
    truncation_length=None,
    output_length=None,
    fade_in=None,
    fade_out=None,
    ir_orders=1,
    inversion_method='regularized',
    **kwargs
):
    if isinstance(inversion_method, str):
        if inversion_method.lower() == 'filtering':
            inversion_method = _filtered_deconvolution
        elif inversion_method.lower() == 'regularized':
            inversion_method = _regularized_deconvolution
        elif inversion_method.lower() == 'analytic':
            inversion_method = _exponential_sweep_analytical_deconvolution
        elif inversion_method.lower() == 'time-reversal':
            inversion_method = _exponential_sweep_time_reversal_deconvolution
        else:
            raise ValueError(f'Unknown inverse filtering method {inversion_method}')

    impulse_response = inversion_method(
        reference_signal=reference_signal,
        measured_signal=measured_signal,
        samplerate=samplerate,
        lower_cutoff=lower_cutoff,
        upper_cutoff=upper_cutoff,
        **kwargs
    )

    if ir_orders == 'whole':
        return impulse_response

    num_samples = impulse_response.shape[-1] // 2

    irs = [impulse_response[..., :num_samples]]
    if ir_orders > 1:
        phase_rate = reference_signal.size / np.log(upper_cutoff / lower_cutoff)
        lags = np.log(np.arange(1, ir_orders + 1)) * phase_rate  # How much each impulse response is before the linear one
        indices = num_samples - np.ceil(lags).astype(int)  # The index where each of the impulse responses start
        for order in range(1, ir_orders):  # Starts with order 1, corresponding to the first harmonic.
            irs.append(impulse_response[indices[order]:indices[order - 1]])

    irs = signal_tools.truncate_signals(irs, length=truncation_length, samplerate=samplerate)
    irs = signal_tools.extend_signals(irs, length=output_length, samplerate=samplerate)
    irs = signal_tools.fade_signals(irs, fade_in=fade_in, fade_out=fade_out, samplerate=samplerate)

    if len(irs) == 1:
        return irs[0]
    return irs


def _filtered_deconvolution(
    reference_signal,
    measured_signal,
    samplerate=None,
    lower_cutoff=None,
    upper_cutoff=None,
    **kwargs
):
    measured_signal = np.asarray(measured_signal)
    num_samples = measured_signal.shape[-1]
    impulse_response = np.fft.irfft(
        np.fft.rfft(measured_signal, 2 * num_samples)
        / np.fft.rfft(reference_signal, 2 * num_samples)
    )

    filter_args = kwargs or {}
    if lower_cutoff is not None and upper_cutoff is not None:
        filter_args.setdefault('Wn', (lower_cutoff, upper_cutoff))
        filter_args['btype'] = 'bandpass'
    elif lower_cutoff is not None:
        filter_args.setdefault('Wn', lower_cutoff)
        filter_args['btype'] = 'highpass'
    elif upper_cutoff is not None:
        filter_args.setdefault('Wn', upper_cutoff)
        filter_args['btype'] = 'lowpass'
    else:
        return impulse_response

    filter_args.setdefault('N', 8)
    filter_args.setdefault('fs', samplerate)
    filter_args.setdefault('ftype', 'butter')
    filter_args['output'] = 'sos'

    sos = scipy.signal.iirfilter(**filter_args)
    impulse_response = scipy.signal.sosfilt(sos, impulse_response)
    return impulse_response


def _regularized_deconvolution(
    reference_signal,
    measured_signal,
    samplerate=None,
    lower_cutoff=None,
    upper_cutoff=None,
    transition_width=1 / 12,
    interior_regularization=0.01,
    exterior_regularization=1,
):
    measured_signal = np.asarray(measured_signal)
    num_samples = measured_signal.shape[-1]

    reference_spectrum = np.fft.rfft(reference_signal, 2 * num_samples)
    samplerate = samplerate or 2  # No given samplerate -> frequencies normalized to Nyquist as in iirfilter
    f = np.fft.rfftfreq(2 * num_samples, 1 / samplerate)
    if lower_cutoff is None:
        lower_cutoff = f[1]
    if upper_cutoff is None:
        upper_cutoff = f[-1]
    lower_cutoff_interior = lower_cutoff * 2**transition_width
    upper_cutoff_interior = upper_cutoff / 2**transition_width

    regularization = np.zeros(reference_spectrum.shape)
    interior_idx = (lower_cutoff_interior <= f) & (f <= upper_cutoff_interior)
    regularization[interior_idx] = 1 / f[interior_idx] * interior_regularization
    regularization[0] = 0

    regularization[f <= lower_cutoff] = exterior_regularization / lower_cutoff
    slope_idx = (lower_cutoff < f) & (f < lower_cutoff_interior)
    regularization[slope_idx] = np.geomspace(exterior_regularization / lower_cutoff, interior_regularization / lower_cutoff_interior, np.sum(slope_idx))

    slope_idx = (upper_cutoff_interior < f) & (f < upper_cutoff)
    regularization[upper_cutoff <= f] = exterior_regularization / upper_cutoff
    regularization[slope_idx] = np.geomspace(interior_regularization / upper_cutoff_interior, exterior_regularization / upper_cutoff, np.sum(slope_idx))

    regularization *= np.mean(np.abs(reference_spectrum[interior_idx])**2 / regularization[interior_idx]) * interior_regularization
    inverse_spectrum = reference_spectrum.conj() / (reference_spectrum.conj() * reference_spectrum + regularization)

    impulse_response = np.fft.irfft(np.fft.rfft(measured_signal, n=2 * num_samples) * inverse_spectrum)
    return impulse_response


def _exponential_sweep_analytical_deconvolution(
    reference_signal,
    measured_signal,
    samplerate,
    lower_cutoff,
    upper_cutoff,
):
    num_samples = reference_signal.size
    phase_rate = num_samples / samplerate / np.log(upper_cutoff / lower_cutoff)
    f = np.fft.rfftfreq(num_samples, 1 / samplerate)
    with np.errstate(divide='ignore', invalid='ignore'):
        inverse_spectrum = 2 * (f / phase_rate)**0.5 * np.exp(-2j * np.pi * f * phase_rate * (1 - np.log(f / lower_cutoff)) + 1j * np.pi / 4)
    inverse_spectrum[0] = 0
    # We have to go back to time to zero pad.
    # It's much easier to generate the correct spectrum with the original reference length.
    inverse_filter = np.fft.irfft(inverse_spectrum) / samplerate

    measured_signal = np.asarray(measured_signal)
    num_samples = measured_signal.shape[-1]

    impulse_response = np.fft.irfft(
        np.fft.rfft(measured_signal, n=2 * num_samples)
        * np.fft.rfft(inverse_filter, n=2 * num_samples),
    )
    impulse_response = np.roll(impulse_response, -reference_signal.size, axis=-1)

    return impulse_response


def _exponential_sweep_time_reversal_deconvolution(
    reference_signal,
    measured_signal,
    lower_cutoff,
    upper_cutoff,
    samplerate=None,
):
    amplitude_correction = 10**(np.linspace(0, -6 * np.log2(upper_cutoff / lower_cutoff), reference_signal.size) / 20)
    inverse_filter = reference_signal[-1::-1] * amplitude_correction

    measured_signal = np.asarray(measured_signal)
    num_samples = measured_signal.shape[-1]

    impulse_response = np.fft.irfft(
        np.fft.rfft(measured_signal, n=2 * num_samples, axis=-1)
        * np.fft.rfft(inverse_filter, n=2 * num_samples),
        axis=-1
    )
    impulse_response = np.roll(impulse_response, -reference_signal.size, axis=-1)

    return impulse_response


def mls_analysis(reference, output):
    output = np.atleast_2d(output)
    seq_len = len(reference)
    order = np.round(np.log2(seq_len + 1)).astype(int)
    reps = int(output.shape[1] / seq_len)

    # Handle different reference mappings
    ref_check = np.sum(reference)
    if ref_check == 1:
        # Mapped with (0, 1) -> (-1, 1)
        reference = (reference + 1) / 2
    elif ref_check == -1:
        # Mapped with (0, 1) -> (1, -1)
        reference = (1 - reference) / 2
    elif ref_check == 2**(order - 1) - 1:
        # Mapped with (0, 1) -> (1, 0)
        reference = 1 - reference

    if np.sum(reference) != 2**(order - 1):
        raise ValueError('MLS reference sequence is not a single iteration of a valid MLS')

    if reps > 1:
        response = output[:, seq_len:reps * seq_len].reshape((-1, reps - 1, seq_len)).mean(axis=1)
    else:
        response = output[:, :seq_len]

    if order <= 8:
        dtype = np.uint8
    elif order <= 16:
        dtype = np.uint16
    elif order <= 32:
        dtype = np.uint32
    elif order <= 64:
        dtype = np.uint64
    else:
        assert False, "You would need at least one zettabyte memory to store this array..."

    ps = np.zeros(seq_len, dtype=dtype)
    idx = np.arange(seq_len)  # Cannot be uint since it will be subtracted to be negative!
    ref = reference.astype(dtype)
    for s in range(order):
        ps[idx] += ref[(idx - s) % seq_len] << (order - 1 - s)

    indices = np.argsort(ps)[2**np.arange(order - 1, -1, -1, dtype=dtype) - 1]  # This also will not be uint.
    pl = np.zeros(seq_len, dtype=dtype)
    for s in range(order):
        pl += ref[(indices[s] - idx) % seq_len] << (order - s - 1)

    transform = np.zeros((output.shape[0], seq_len + 1))
    transform[:, ps] = response
    for _ in range(order):
        transform = np.concatenate((transform[:, ::2] + transform[:, 1::2], transform[:, ::2] - transform[:, 1::2]), axis=1)
    ir = transform[:, pl] / (seq_len + 1)
    return np.squeeze(ir)
